/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#define BGPATH PACKAGE_DATA_DIR"/"

#ifdef TESTING
#undef PACKAGE_DATA_DIR
#undef BGPATH
#define PACKAGE_DATA_DIR "../data"
#define BGPATH "../data/"
#endif

#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <signal.h>
#include <unicode.h>

#include <gtk/gtk.h>
#include <gnome.h>
#include <glade/glade.h>
#include <glade/glade-build.h>

#ifdef WITH_GNOME
#  include <libgnomeui/libgnomeui.h>
#endif

#include "timidity-io.h"
#include "track.h"

/* search for a widget given a known widget (in the same widget tree) 
 * using libglade.
 */
static GtkWidget *lookup_widget(GtkWidget * widget,
				const gchar * widget_name)
{
	GladeXML *xml;
	GtkWidget *wdgt;

	xml = glade_get_widget_tree(widget);
	wdgt = glade_xml_get_widget(xml, widget_name);


	if (wdgt == NULL)
		g_warning("Widget not found: %s", widget_name);

	return wdgt;
}

const static gchar *GKaraokeGladeFile = PACKAGE_DATA_DIR "/gkaraoke.glade";
static gchar *lyricsfont;
static const gchar lyricsfont_default[] = "Sans Bold Italic 32";

struct stAppData {
	GtkWidget *mainwindow;
	gint timidity_pid;
	gint timidity_connected;
	gint tries_of_timidity_connection;
	struct stSequencer *sequencer;
	struct stCurrentFile *current_file;
	struct stPlayHandle *playhandle;
	gchar *next_file;
};

void on_open1_activate(GtkWidget * window);
void on_about1_activate(GtkWidget * window);
void on_fileentry1_activate(GtkWidget * window);
void on_preferences1_activate(GtkWidget * window);
void on_lyricdelay_value_changed(GtkWidget * window);
void on_preferences1_cancel(GtkWidget * window);
void on_preferences1_ok(GtkWidget * window);
void on_ok_filesel_clicked(GtkWidget * window);
void on_lyricfont_font_set(GtkWidget * window);

void on_about1_activate(GtkWidget * window)
{
	GtkWidget *about;
	const gchar *authors[] = {
		"Fabio Ferrari <fabio@floripa.com.br>",
		NULL
	};
	const gchar *documenters[] = {
		NULL
	};
	const gchar *translator = "Fabio Ferrari <fabio@floripa.com.br>";

	about = gnome_about_new("GKaraoke", VERSION,
				"(c)2003 Fabio Ferrari",
				_("The Gnome Karaoke Simulator"),
				authors, documenters, translator, NULL);
	gtk_widget_show(about);
}

static void open_new_kar_file(gchar * filename, struct stAppData *appdata);

void on_ok_filesel_clicked(GtkWidget * window)
{
	struct stAppData *appdata;
	appdata = gtk_object_get_data(GTK_OBJECT(window), "AppData");
	open_new_kar_file((gchar *) gtk_file_selection_get_filename
			  (GTK_FILE_SELECTION(window)), appdata);
	gtk_widget_destroy(window);
}

void on_open1_activate(GtkWidget * window)
{
	struct stAppData *appdata;
	GladeXML *gladedata;
	GtkWidget *win;
	gchar *rootnode = "OpenKarFile";
	appdata = gtk_object_get_data(GTK_OBJECT(window), "AppData");
	/* construct the interface */
	gladedata = glade_xml_new(GKaraokeGladeFile, rootnode, NULL);
	if (!gladedata) {
		g_error
		    (_
		     ("Could not read interface description (glade) %s from file %s"),
		     rootnode, GKaraokeGladeFile);
	}
	glade_xml_signal_autoconnect(gladedata);
	win = glade_xml_get_widget(gladedata, rootnode);
	g_object_unref(G_OBJECT(gladedata));

	gtk_object_set_data(GTK_OBJECT(win), "AppData", appdata);
}

void on_fileentry1_activate(GtkWidget * window)
{
	struct stAppData *appdata;
	GnomeFileEntry *fileentry;
	appdata = gtk_object_get_data(GTK_OBJECT(window), "AppData");
	fileentry = GNOME_FILE_ENTRY(lookup_widget(window, "fileentry1"));
	open_new_kar_file(gnome_file_entry_get_full_path(fileentry, FALSE),
			  appdata);
}

static void change_lyric_font(const gchar * newfont);

void on_preferences1_activate(GtkWidget * window)
{
	struct stAppData *appdata;
	GladeXML *gladedata;
	GtkWidget *prefwindow, *lyricdelay, *lyricfont;
	gchar *rootnode = "PreferencesDialog";
	gchar *prev_font;

	appdata = gtk_object_get_data(GTK_OBJECT(window), "AppData");
	/* construct the interface */
	gladedata = glade_xml_new(GKaraokeGladeFile, rootnode, NULL);
	if (!gladedata) {
		g_error
		    (_
		     ("Could not read interface description (glade) %s from file %s"),
		     rootnode, GKaraokeGladeFile);
	}
	glade_xml_signal_autoconnect(gladedata);
	prefwindow = glade_xml_get_widget(gladedata, rootnode);
	g_signal_connect_swapped(GTK_OBJECT(prefwindow), "destroy",
				 G_CALLBACK(g_object_unref), gladedata);

	lyricdelay = glade_xml_get_widget(gladedata, "lyricdelay");
	lyricfont = glade_xml_get_widget(gladedata, "lyricfont");
	gtk_range_set_value(GTK_RANGE(lyricdelay),
			    (gdouble) track_get_lyric_delay());
	gtk_object_set_data(GTK_OBJECT(prefwindow), "prev_delay",
			    GUINT_TO_POINTER(track_get_lyric_delay()));
	gnome_font_picker_set_font_name(GNOME_FONT_PICKER(lyricfont),
					lyricsfont);
	prev_font = g_strdup(lyricsfont);
	gtk_object_set_data_full(GTK_OBJECT(prefwindow), "prev_font",
				 prev_font, g_free);
}

void on_lyricdelay_value_changed(GtkWidget * window)
{
	GtkWidget *lyricdelay;
	lyricdelay = lookup_widget(window, "lyricdelay");
	track_set_lyric_delay((guint)
			      gtk_range_get_value(GTK_RANGE(lyricdelay)));
}

void on_lyricfont_font_set(GtkWidget * window)
{
	GtkWidget *lyricfont;
	lyricfont = lookup_widget(window, "lyricfont");
	change_lyric_font(gnome_font_picker_get_font_name
			  (GNOME_FONT_PICKER(lyricfont)));
}

void on_preferences1_cancel(GtkWidget * window)
{
	guint prevdelay;
	gchar *prev_font;
	prevdelay =
	    GPOINTER_TO_UINT(gtk_object_get_data
			     (GTK_OBJECT(window), "prev_delay"));
	track_set_lyric_delay(prevdelay);
	prev_font = gtk_object_get_data(GTK_OBJECT(window), "prev_font");
	change_lyric_font(prev_font);
	gtk_widget_destroy(window);
}

void on_preferences1_ok(GtkWidget * window)
{
	gtk_widget_destroy(window);
	gnome_config_set_string("/GKaraoke/General/lyric_font",
				lyricsfont);
	gnome_config_set_int("/GKaraoke/General/lyric_delay",
			     track_get_lyric_delay());
	gnome_config_sync();
}

static gint start_timidity(void)
{
	gint pid;
	pid = fork();
	if (pid < 0)
		return -1;

	if (pid == 0) {
		gint open_max;
		gint i;

		open_max = sysconf(_SC_OPEN_MAX);
		for (i = 3; i < open_max; i++)
			close(i);

		execlp("timidity", "timidity", "-ir", "7777", NULL);
		exit(-1);
	}
	return pid;
}

static gint connecting_to_timidity(gpointer data);
static void timidity_connection_result(gpointer p_appdata,
				       struct stSequencer *sequencer);

#define CANVAS_WIDTH 640
#define CANVAS_HEIGHT 480

static void reposition_lyric_lines(void);

static double canvas_scale = 1.0;

static void canvas_new_size_cb(GtkObject * object, GtkAllocation * newsize)
{
	GtkWidget *canvas1 = GTK_WIDGET(object);
	if (CANVAS_WIDTH * newsize->height >
	    CANVAS_HEIGHT * newsize->width) {
		canvas_scale = (double) newsize->width / CANVAS_WIDTH;
	} else {
		canvas_scale = (double) newsize->height / CANVAS_HEIGHT;
	}
	gnome_canvas_set_pixels_per_unit(GNOME_CANVAS(canvas1),
					 canvas_scale);
	reposition_lyric_lines();
}

static GnomeCanvasItem *canvastextitem_left[2], *canvastextitem_right[2];
static GnomeCanvasItem *canvasimageitem;

static gchar *backgroundimages[] = {
	BGPATH "background1.jpg",
	BGPATH "background2.jpg",
	BGPATH "background3.jpg",
	NULL
};

static GdkPixbuf **background_images_pixbuf;
static gint background_images_cont;

static void destroy_pixbuf_image_cb(GtkObject * object,
				    GdkEventAny * event)
{
	gint ct;
	for (ct = 0; background_images_pixbuf[ct] != NULL; ct++)
		gdk_pixbuf_unref(background_images_pixbuf[ct]);
	g_free(background_images_pixbuf);
}

static void create_background_images(GnomeCanvasGroup * group)
{
	gint ct, qt, ps;
	for (qt = 0; backgroundimages[qt] != NULL; qt++);
	background_images_pixbuf = g_new0(GdkPixbuf *, qt + 1);
	for (ct = 0, ps = 0; ct < qt; ct++) {
		GError *error = NULL;
		background_images_pixbuf[ps] =
		    gdk_pixbuf_new_from_file(backgroundimages[ct], &error);
		if (background_images_pixbuf[ps] != NULL)
			ps++;
	}

	background_images_cont = 0;
	canvasimageitem =
	    gnome_canvas_item_new(group, gnome_canvas_pixbuf_get_type(),
				  "pixbuf",
				  background_images_pixbuf
				  [background_images_cont], "x",
				  (double) 0, "y", (double) 0, "anchor",
				  GTK_ANCHOR_NW,
				  "width-set", (gboolean) TRUE,
				  "height-set", (gboolean) TRUE,
				  "width", (double) CANVAS_WIDTH,
				  "height", (double) CANVAS_HEIGHT, NULL);
	/* Clean up gdk_pixbuf when Canvas item is destroyed */
	gtk_signal_connect(GTK_OBJECT(canvasimageitem), "destroy",
			   GTK_SIGNAL_FUNC(destroy_pixbuf_image_cb), NULL);
}

static void change_background_image(void)
{
	background_images_cont++;
	if (background_images_pixbuf[background_images_cont] == NULL) {
		background_images_cont = 0;
	}
	gnome_canvas_item_set(canvasimageitem, "pixbuf",
			      background_images_pixbuf
			      [background_images_cont], NULL);
}

static gboolean canvastextactive = FALSE;

static void define_canvas_text_desactivaded(GtkWidget * window)
{
	canvastextactive = FALSE;
}

static void create_lines_text(GnomeCanvasGroup * group)
{
	canvastextitem_left[0] =
	    gnome_canvas_item_new(group, GNOME_TYPE_CANVAS_TEXT, "x",
				  320.0, "y", 460.0, "text", "",
				  "justification", GTK_JUSTIFY_CENTER,
				  "anchor", GTK_ANCHOR_SOUTH, "fill_color",
				  "green", "font", lyricsfont, NULL);
	canvastextitem_right[0] =
	    gnome_canvas_item_new(group, GNOME_TYPE_CANVAS_TEXT, "x",
				  320.0, "y", 460.0, "text", "",
				  "justification", GTK_JUSTIFY_CENTER,
				  "anchor", GTK_ANCHOR_SOUTH, "fill_color",
				  "red", "font", lyricsfont, NULL);
	canvastextitem_left[1] =
	    gnome_canvas_item_new(group, GNOME_TYPE_CANVAS_TEXT, "x",
				  320.0, "y", 460.0, "text", "",
				  "justification", GTK_JUSTIFY_CENTER,
				  "anchor", GTK_ANCHOR_SOUTH, "fill_color",
				  "green", "font", lyricsfont, NULL);
	canvastextitem_right[1] =
	    gnome_canvas_item_new(group, GNOME_TYPE_CANVAS_TEXT, "x",
				  320.0, "y", 460.0, "text", "",
				  "justification", GTK_JUSTIFY_CENTER,
				  "anchor", GTK_ANCHOR_SOUTH, "fill_color",
				  "red", "font", lyricsfont, NULL);
	gtk_signal_connect(GTK_OBJECT(canvastextitem_right[1]),
			   "destroy",
			   GTK_SIGNAL_FUNC
			   (define_canvas_text_desactivaded), NULL);
	canvastextactive = TRUE;
}

static gchar *convert_to_unicode(gchar * line)
{
	unicode_iconv_t ic;
	gchar *inbuf, *resultstr;
	int inlen;
	GByteArray *result;

	result = g_byte_array_new();
	ic = unicode_iconv_open("utf-8", "iso-8859-1");
	inbuf = line;
	inlen = strlen(line);
	while (inlen) {
		gchar outbase[100], *outbuf = outbase;
		gint outlen = 100;
		if (unicode_iconv
		    (ic, (const char **) &inbuf, &inlen, &outbuf,
		     &outlen) < 0)
			break;
		g_byte_array_append(result, outbase, outbuf - outbase);
	}
	unicode_iconv_close(ic);
	g_byte_array_append(result, "\0", 1);
	resultstr = result->data;
	g_byte_array_free(result, FALSE);
	return resultstr;
}

static void define_lyric_line(gint line, gchar * left, gchar * right)
{
	gchar *left_utf8, *right_utf8;

	if (canvastextactive == FALSE)
		return;

	left_utf8 = convert_to_unicode(left);
	right_utf8 = convert_to_unicode(right);
	gnome_canvas_item_set(canvastextitem_left[line], "text", left_utf8,
			      NULL);
	gnome_canvas_item_set(canvastextitem_right[line], "text",
			      right_utf8, NULL);
	reposition_lyric_lines();
	g_free(left_utf8);
	g_free(right_utf8);
}

static void change_lyric_font(const gchar * newfont)
{
	gint line;
	g_free(lyricsfont);
	lyricsfont = g_strdup(newfont);
	for (line = 0; line < 2; line++) {
		gnome_canvas_item_set(canvastextitem_left[line], "font",
				      lyricsfont, NULL);
		gnome_canvas_item_set(canvastextitem_right[line], "font",
				      lyricsfont, NULL);
	}
	reposition_lyric_lines();
}

static void reposition_lyric_lines(void)
{
	double sizeleft, sizeright, posleft, posright;
	double height;
	gint line;
	for (line = 0; line < 2; line++) {
		gtk_object_get(GTK_OBJECT(canvastextitem_left[line]),
			       "text_width", &sizeleft, NULL);
		gtk_object_get(GTK_OBJECT(canvastextitem_right[line]),
			       "text_width", &sizeright, NULL);
		posleft = -sizeright / 2;
		posright = sizeleft / 2;
		gnome_canvas_item_set(canvastextitem_left[line],
				      "x_offset", posleft, NULL);
		gnome_canvas_item_set(canvastextitem_right[line],
				      "x_offset", posright, NULL);
	}

	gtk_object_get(GTK_OBJECT(canvastextitem_left[1]),
		       "text_height", &height, NULL);
	gnome_canvas_item_set(canvastextitem_left[0],
			      "y_offset", -height * 1.2, NULL);
	gnome_canvas_item_set(canvastextitem_right[0],
			      "y_offset", -height * 1.2, NULL);
}

static GtkWidget *create_gkaraoke_window_from_glade(struct stAppData
						    *appdata)
{
	GladeXML *gladedata;
	const char *rootnode = "GKaraokeApp";
	GtkWidget *win;

	/* construct the interface */
	gladedata = glade_xml_new(GKaraokeGladeFile, rootnode, NULL);
	if (!gladedata) {
		g_error
		    (_
		     ("Could not read interface description (glade) %s from file %s"),
		     rootnode, GKaraokeGladeFile);
	}

	/* look up handler names in the symbol table and automatically
	 * connect the signals. This requires that your program export
	 * dynamic symbols.  If you link using libtool, passing the
	 * -export-dynamic link flag will do this in a cross platform
	 * manner.  For GNU ld, the --export-dynamic argument should
	 * work.
	 *
	 * If you don't want to do symbol lookups like this, the
	 * glade_xml_signal_connect() function allows you to connect
	 * individual handlers up. */
	glade_xml_signal_autoconnect(gladedata);

	win = glade_xml_get_widget(gladedata, rootnode);

	/* preserve glade structure until the end of the window */
	g_signal_connect_swapped(GTK_OBJECT(win), "destroy",
				 G_CALLBACK(g_object_unref), gladedata);

	return win;
}

int main(int argc, char *argv[])
{
	GtkWidget *mainwindow, *appbar1, *canvas1;
	GnomeCanvasGroup *group;
	struct stAppData *appdata;
	gchar *buf;

#ifdef ENABLE_NLS
	bindtextdomain(PACKAGE, PACKAGE_LOCALE_DIR);
	textdomain(PACKAGE);
#endif

	/* initialise libraries */
	gtk_init(&argc, &argv);
	if (!gnome_program_init("GKaraoke", VERSION,
				LIBGNOMEUI_MODULE, argc, argv, NULL))
		g_error("Cannot gnome_program_init ()");

	buf =
	    g_strdup_printf("/GKaraoke/General/lyric_font=%s",
			    lyricsfont_default);
	lyricsfont = gnome_config_get_string(buf);
	g_free(buf);

	buf =
	    g_strdup_printf("/GKaraoke/General/lyric_delay=%d",
			    track_get_lyric_delay());
	track_set_lyric_delay((guint) gnome_config_get_int(buf));
	g_free(buf);

	appdata = g_new0(struct stAppData, 1);

	if (argc == 2) {
		appdata->next_file = g_strdup(argv[1]);
	}

	mainwindow = create_gkaraoke_window_from_glade(appdata);
	gtk_object_set_data(GTK_OBJECT(mainwindow), "AppData", appdata);
	appdata->mainwindow = mainwindow;

	canvas1 = lookup_widget(mainwindow, "canvas1");
	gtk_signal_connect(GTK_OBJECT(canvas1), "size_allocate",
			   GTK_SIGNAL_FUNC(canvas_new_size_cb), NULL);
	group = gnome_canvas_root(GNOME_CANVAS(canvas1));
	create_background_images(group);
	create_lines_text(group);

	appbar1 = lookup_widget(mainwindow, "appbar1");
	gnome_appbar_set_default(GNOME_APPBAR(appbar1), "GKaraoke");
	gnome_appbar_push(GNOME_APPBAR(appbar1), _("Starting Timidity"));
	appdata->timidity_pid = start_timidity();
	if (appdata->timidity_pid < 0) {
		GtkWidget *mbox;
		mbox =
		    gnome_error_dialog
		    (_("Unable to fork process to start Timidity."));
		gtk_signal_connect(GTK_OBJECT(mbox), "destroy",
				   GTK_SIGNAL_FUNC(gtk_main_quit), NULL);
		gtk_widget_show(mbox);
	}
	if (appdata->timidity_pid >= 0) {
		gnome_appbar_pop(GNOME_APPBAR(appbar1));
		gnome_appbar_push(GNOME_APPBAR(appbar1),
				  _("Connecting to Timidity"));
		gtk_timeout_add(250, connecting_to_timidity, appdata);
	}
	gnome_appbar_pop(GNOME_APPBAR(appbar1));

	gtk_main();

	open_new_kar_file(NULL, appdata);
	if (appdata->timidity_connected)
		close_sequencer(appdata->sequencer);
	if (appdata->timidity_pid >= 0)
		kill(appdata->timidity_pid, SIGTERM);
	g_free(appdata);
	return 0;
}

static gint connecting_to_timidity(gpointer data)
{
	struct stAppData *appdata;
	appdata = (struct stAppData *) data;
	open_sequencer(appdata, timidity_connection_result);
	return FALSE;
}

static void timidity_connection_result(gpointer p_appdata,
				       struct stSequencer *sequencer)
{
	GtkWidget *appbar1;
	struct stAppData *appdata = (struct stAppData *) p_appdata;
	if (sequencer == NULL) {
		GtkWidget *mbox;
		appdata->tries_of_timidity_connection++;
		if (appdata->tries_of_timidity_connection <= 3000 / 250) {
			gtk_timeout_add(250, connecting_to_timidity,
					appdata);
			return;
		}
		mbox =
		    gnome_error_dialog(_
				       ("Unable to connect to Timidity."));
		gtk_window_set_modal(GTK_WINDOW(mbox), TRUE);
		gtk_signal_connect(GTK_OBJECT(mbox), "destroy",
				   GTK_SIGNAL_FUNC(gtk_main_quit),
				   appdata);
		gtk_widget_show(mbox);
		return;
	}
	appdata->sequencer = sequencer;
	appdata->timidity_connected = TRUE;
	appbar1 = lookup_widget(appdata->mainwindow, "appbar1");
	gnome_appbar_refresh(GNOME_APPBAR(appbar1));

	if (appdata->next_file != NULL) {
		open_new_kar_file(appdata->next_file, appdata);
		g_free(appdata->next_file);
		appdata->next_file = NULL;
	}
}

static void open_new_kar_file(gchar * filename, struct stAppData *appdata)
{
	GtkWidget *appbar1;
	gchar *msg;

	if (appdata->playhandle != NULL) {
		track_clear_playfile(appdata->playhandle);
		appdata->playhandle = NULL;
	}
	if (appdata->current_file != NULL) {
		clear_tracks(appdata->current_file);
		appdata->current_file = NULL;
	}

	if (filename == NULL)
		return;
	if (strlen(filename) == 0)
		return;

	appdata->current_file = load_tracks(filename);
	if (appdata->current_file == NULL) {
		GtkWidget *mbox;
		msg = g_strdup_printf(_("Unable to read: %s"), filename);
		mbox = gnome_error_dialog(msg);
		gtk_window_set_modal(GTK_WINDOW(mbox), TRUE);
		g_free(msg);
		return;
	}
	appbar1 = lookup_widget(appdata->mainwindow, "appbar1");
	msg = g_strdup_printf(_("GKaraoke: %s."), filename);
	gnome_appbar_set_default(GNOME_APPBAR(appbar1), msg);
	g_free(msg);
	appdata->playhandle =
	    track_playfile(appdata->sequencer,
			   appdata->current_file, define_lyric_line,
			   change_background_image);
}
