/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/syscall.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netdb.h>		/* for gethostbyname */
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdarg.h>
#include <stdlib.h>
#include <unistd.h>

/* Please modify include path if need */
#include <sys/soundcard.h>
#include <linux/awe_voice.h>
#include <errno.h>

#include <glib.h>
#include <gdk/gdk.h>

#include "my_midi.h"
#include "timidity-io.h"

#define TIMIDITY_LOW_DELAY

#ifdef TIMIDITY_LOW_DELAY
#define BUF_LOW_SYNC	0.1
#define BUF_HIGH_SYNC	0.15
#else
#define BUF_LOW_SYNC	0.4
#define BUF_HIGH_SYNC	0.8
#endif				/* TIMIDITY_ASYNC_MODE */

/* Default port */
#define TIMIDITY_CONTROL_PORT 7777

#ifndef INADDR_NONE
#define INADDR_NONE 0xffffffff	/* -1 return */
#endif				/* INADDR_NONE */

struct fd_read_buffer {
	char buff[BUFSIZ];
	/* count: beginning of read pointer
	 * size:  end of read pointer
	 * fd:    file descripter for input
	 */
	int count, size;
};

struct stSequencer {
	gint ctrl_fd, data_fd;
	gint gdk_input_tag;
	gpointer appdata;
	GByteArray *read_buffer;

	GHashTable *noteactive;

	double counter_ms;
	void (*sync_reply_cb) (double ms_avail, gpointer appdata);
	gpointer sync_reply_cb_data;

	struct fd_read_buffer control_fd;
};

static int fdgets(struct stSequencer *sequencer,
		  char *buff, size_t buff_size, struct fd_read_buffer *p);
static char *timidity_ctl_command(struct stSequencer *sequencer,
				  const char *fmt, ...);

static int timidity_timebase;

#define seqbuf_dump() sequencer_dumpbuf(sequencer)

SEQ_DEFINEBUF(128);

SEQ_USE_EXTBUF();
void sequencer_dumpbuf(struct stSequencer *sequencer)
{
	if (_seqbufptr)
		if (write(sequencer->data_fd, _seqbuf, _seqbufptr) == -1) {
			perror("write timidity");
			exit(-1);
		}
	_seqbufptr = 0;
}

static void timidity_meta_seq(struct stSequencer *sequencer,
			      int p1, int p2, int p3)
{
	_CHN_COMMON(0, 0xff, 0x7f, p1, p2, p3);
	SEQ_DUMPBUF();
}

/* Return internet IP address in network byte order */
static unsigned int host_ip_address(const char *address)
{
	unsigned int addr;
	struct hostent *hp;

	if ((addr = inet_addr(address)) != INADDR_NONE)
		return addr;

	if ((hp = gethostbyname(address)) == NULL) {
		g_warning("Unknown host name: %s", address);
		exit(1);
	}
	memcpy(&addr, hp->h_addr, hp->h_length);
	return addr;
}

static int connect_to_server(const char *hostname, unsigned short tcp_port)
{
	int fd, len;
	struct sockaddr_in in;
	unsigned int addr;
	struct hostent *hp;

	if ((fd = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
		perror("socket");
		exit(1);
	}

	memset(&in, 0, sizeof(in));
	in.sin_family = AF_INET;
	in.sin_port = htons(tcp_port);
	addr = host_ip_address(hostname);
	memcpy(&in.sin_addr, &addr, 4);

	hp = gethostbyaddr((const char *) &in.sin_addr, 4, AF_INET);
	if (hp != NULL)
		g_message("Try to connect to %s (addr=%s, port=%d)",
			  hp->h_name, inet_ntoa(in.sin_addr), tcp_port);
	else
		g_message("Try to connect to %s (port=%d)",
			  inet_ntoa(in.sin_addr), tcp_port);
	alarm(15);
	if (connect(fd, (struct sockaddr *) &in, sizeof(in)) < 0) {
		g_warning("connect: %s", strerror(errno));
		return -1;
	}
	alarm(0);

	len = sizeof(in);
	if (getsockname(fd, (struct sockaddr *) &in, &len) < 0)
		g_warning("getsockname: %s", strerror(errno));
	return fd;
}

static int fdgets(struct stSequencer *sequencer,
		  char *buff, size_t buff_size, struct fd_read_buffer *p)
{
	int n, len, count, size;
	char *buff_endp = buff + buff_size - 1, *pbuff, *beg;

	if (buff_size == 0)
		return 0;
	else if (buff_size == 1) {	/* buff == buff_endp */
		*buff = '\0';
		return 0;
	}

	len = 0;
	count = p->count;
	size = p->size;
	pbuff = p->buff;
	beg = buff;
	do {
		if (count == size) {
			alarm(10);
			if ((n =
			     read(sequencer->ctrl_fd, pbuff,
				  BUFSIZ)) <= 0) {
				alarm(0);
				*buff = '\0';
				if (n == 0) {
					p->count = p->size = 0;
					return buff - beg;
				}
				return -1;	/* < 0 error */
			}
			alarm(0);
			count = p->count = 0;
			size = p->size = n;
		}
		*buff++ = pbuff[count++];
	} while (*(buff - 1) != '\n' && buff != buff_endp);
	*buff = '\0';
	p->count = count;
	return buff - beg;
}

static char *timidity_ctl_command(struct stSequencer *sequencer,
				  const char *fmt, ...)
{
	static char buff[BUFSIZ];
	int status, len;
	va_list ap;

	if (fmt != NULL) {
		va_start(ap, fmt);
		vsprintf(buff, fmt, ap);
		va_end(ap);
		len = strlen(buff);
		if (len > 0 && buff[len - 1] != '\n')
			buff[len++] = '\n';
		write(sequencer->ctrl_fd, buff, len);
	}

	while (1) {
		if (fdgets
		    (sequencer, buff, sizeof(buff),
		     &sequencer->control_fd) <= 0) {
			strcpy(buff, "Read error\n");
			break;
		}
		status = atoi(buff);
		if (400 <= status && status <= 499) {	/* Error of data stream */
			fputs(buff, stderr);
			continue;
		}
		break;
	}
	return buff;
}

static inline void timidity_request_sync(struct stSequencer *sequencer)
{
	/* send SYNC request */
	timidity_meta_seq(sequencer, 0x02, 0x00, 0);
}

static void stop_pending_midi_notes(struct stSequencer *sequencer);

void timidity_set_sync_cb(struct stSequencer *sequencer,
			  void (*sync_reply_cb) (double ms_avail,
						 gpointer appdata),
			  gpointer appdata)
{
	sequencer->sync_reply_cb = sync_reply_cb;
	sequencer->sync_reply_cb_data = appdata;

	stop_pending_midi_notes(sequencer);
}

static const double blockmidisize = 200.0;	/* 200 ms */

void sequencer_wait(struct stSequencer *sequencer, double milisec)
{
	sequencer->counter_ms += milisec;
	SEQ_WAIT_TIME((guint)
		      ((sequencer->counter_ms * timidity_timebase) /
		       1000.0));
}

#define MIDIDEVNUM 0

static inline void memorize_midi_event(struct stSequencer *sequencer,
				       guchar status, guchar byte1,
				       guchar byte2)
{
	gpointer key;
	key = (gpointer) (((status & 0xFF) << 8) | (byte1 & 0xFF));
	if (byte2 > 0)
		g_hash_table_insert(sequencer->noteactive, key,
				    (gpointer) TRUE);
	else
		g_hash_table_remove(sequencer->noteactive, key);
}

void sequencer_send_midi_event(struct stSequencer *sequencer,
			       guchar status, guchar data1, guchar data2)
{
	/* I hereby abolish all foul  SYSEX from this program   */
	if (status == 0xf0)
		return;

	switch (status & 0xf0) {
	case NOTE_ON:
	case NOTE_OFF:
	case KEY_PRESSURE:
		memorize_midi_event(sequencer, status, data1, data2);
	case PITCH_BEND:
	case CONTROL_CHANGE:
		SEQ_MIDIOUT(MIDIDEVNUM, status);
		SEQ_MIDIOUT(MIDIDEVNUM, data1);
		SEQ_MIDIOUT(MIDIDEVNUM, data2);
		break;
	case PROGRAM_CHANGE:
	case CHANNEL_PRESSURE:
		SEQ_MIDIOUT(MIDIDEVNUM, status);
		SEQ_MIDIOUT(MIDIDEVNUM, data1);
		break;
	}
	SEQ_DUMPBUF();
}

static gboolean send_stop_midi_event(guint key, gpointer value,
				     struct stSequencer *sequencer)
{
	guchar status, data1, data2 = 0;

	status = (key & 0xFF00) >> 8;
	data1 = key & 0xFF;

	SEQ_MIDIOUT(MIDIDEVNUM, status);
	SEQ_MIDIOUT(MIDIDEVNUM, data1);
	SEQ_MIDIOUT(MIDIDEVNUM, data2);
	SEQ_DUMPBUF();
	return TRUE;
}

static void stop_pending_midi_notes(struct stSequencer *sequencer)
{
	g_hash_table_foreach_remove(sequencer->noteactive,
				    (GHRFunc) send_stop_midi_event,
				    sequencer);
}

static void sequencer_process_reply(struct stSequencer *sequencer,
				    gchar * reply)
{
	int status;

	status = atoi(reply);
	if (400 <= status && status <= 499) {	/* Error of data stream */
		g_warning("timidity reply: %s", reply);
		return;
	}

	if (status == 301) {
		if (sequencer->sync_reply_cb != NULL) {
			(*sequencer->sync_reply_cb) (blockmidisize,
						     sequencer->
						     sync_reply_cb_data);
		} else {
			sequencer_wait(sequencer, blockmidisize);
		}
		timidity_request_sync(sequencer);
	}
}

static void sequencer_ctrl_reply(gpointer data, gint source_fd,
				 GdkInputCondition condition)
{
	struct stSequencer *sequencer = data;
	gchar buff[BUFSIZ];
	gint len;

	len = read(sequencer->ctrl_fd, buff, sizeof(buff));
	if (len <= 0) {
		g_error("Lost connection to timidity!");
		exit(-1);
	}
	g_byte_array_append(sequencer->read_buffer, buff, len);

	for (;;) {
		gchar *pos, *reply;
		gint replysz, whatisleftsz;
		/* search for a new line */
		pos =
		    memchr(sequencer->read_buffer->data, '\n',
			   sequencer->read_buffer->len);
		if (pos == NULL)
			break;

		/* found at least one complete reply */
		pos++;		/* newline should be part of the reply */
		replysz = pos - (gchar *) sequencer->read_buffer->data;

		/* make a copy of the line read without newline */
		reply =
		    g_strndup(sequencer->read_buffer->data, replysz - 1);

		/* shift what is left over the line read */
		whatisleftsz = sequencer->read_buffer->len - replysz;
		if (whatisleftsz)
			g_memmove(sequencer->read_buffer->data, pos,
				  whatisleftsz);
		g_byte_array_set_size(sequencer->read_buffer,
				      whatisleftsz);

		sequencer_process_reply(sequencer, reply);
		g_free(reply);
	}
}

void open_sequencer(gpointer appdata,
		    void (*timidity_connection_result) (gpointer appdata,
							struct stSequencer
							* sequencer))
{
	struct stSequencer *sequencer;
	char *res;
	int i, data_port;
	static char timidity_host[] = "127.0.0.1";
	static int timidity_port = TIMIDITY_CONTROL_PORT;

	sequencer = g_new0(struct stSequencer, 1);
	sequencer->appdata = appdata;
	sequencer->read_buffer = g_byte_array_new();

	sequencer->ctrl_fd =
	    connect_to_server(timidity_host, timidity_port);
	res = timidity_ctl_command(sequencer, NULL);
	if (atoi(res) != 220) {
		g_message("Can't connect timidity: (host=%s, port=%d)",
			  timidity_host, timidity_port);
		g_byte_array_free(sequencer->read_buffer, TRUE);
		g_free(sequencer);
		(*timidity_connection_result) (appdata, NULL);
		return;
	}
	fputs(res, stdout);

	res = timidity_ctl_command(sequencer, "SETBUF %f %f",
				   BUF_LOW_SYNC, BUF_HIGH_SYNC);
	if (atoi(res) != 200)
		g_warning("WARNING: %s", res);

	i = 1;
	if (*(char *) &i == 1)
		res = timidity_ctl_command(sequencer, "OPEN lsb");
	else
		res = timidity_ctl_command(sequencer, "OPEN msb");

	if (atoi(res) != 200) {
		g_warning("Can't open timidity data connection: %s", res);
		g_byte_array_free(sequencer->read_buffer, TRUE);
		close(sequencer->ctrl_fd);
		g_free(sequencer);
		(*timidity_connection_result) (appdata, NULL);
		return;
	}

	data_port = atoi(res + 4);
	sequencer->data_fd = connect_to_server(timidity_host, data_port);
	res = timidity_ctl_command(sequencer, NULL);
	if (atoi(res) != 200) {
		g_warning("Can't connect timidity: %s\t(host=%s, port=%d)",
			  res, timidity_host, data_port);
		g_byte_array_free(sequencer->read_buffer, TRUE);
		close(sequencer->ctrl_fd);
		g_free(sequencer);
		(*timidity_connection_result) (appdata, NULL);
		return;
	}
	res = timidity_ctl_command(sequencer, "TIMEBASE");
	if (atoi(res) != 200) {
		g_warning("Can't find timebase from Timidity");
		g_byte_array_free(sequencer->read_buffer, TRUE);
		close(sequencer->data_fd);
		close(sequencer->ctrl_fd);
		g_free(sequencer);
		(*timidity_connection_result) (appdata, NULL);
		return;
	}
	timidity_timebase = atoi(res + 4);
	sequencer->gdk_input_tag = gdk_input_add(sequencer->ctrl_fd,
						 GDK_INPUT_READ,
						 sequencer_ctrl_reply,
						 sequencer);

	SEQ_START_TIMER();
	/* SEQ_SET_TEMPO(360); <- what is that for timidity? */

	sequencer->counter_ms = 0.0;
	/* send two sync to timidity, starting ping-pong with ev. blocks */
	timidity_request_sync(sequencer);
	sequencer_wait(sequencer, blockmidisize);
	timidity_request_sync(sequencer);

	sequencer->noteactive =
	    g_hash_table_new(g_direct_hash, g_direct_equal);

	(*timidity_connection_result) (appdata, sequencer);
}

int close_sequencer(struct stSequencer *sequencer)
{
	gdk_input_remove(sequencer->gdk_input_tag);
	close(sequencer->data_fd);
	close(sequencer->ctrl_fd);
	g_byte_array_free(sequencer->read_buffer, TRUE);
	g_hash_table_destroy(sequencer->noteactive);
	g_free(sequencer);
	return 0;
}

#if 0
static int timidity_eot(struct stSequencer *sequencer)
{
	timidity_meta_seq(sequencer, 0x00, 0x00, 0);	/* End of playing */
	return timidity_sync(sequencer, 0);
}

int finish_and_close_sequencer(struct stSequencer *sequencer)
{
	char *res;

	timidity_eot(sequencer);
	timidity_sync(sequencer, 0);
	do {
		res = timidity_ctl_command(sequencer, "QUIT");
	} while (atoi(res) && atoi(res) != 302);
	return close_sequencer(sequencer);
}
#endif

#if 0
static int timidity_sync(struct stSequencer *sequencer, int centsec)
{
	char *res;
	int status;
	unsigned long sleep_usec;

	timidity_meta_seq(sequencer, 0x02, 0x00, centsec);	/* Wait playout */

	/* Wait "301 Sync OK" */
	do {
		res = timidity_ctl_command(sequencer, NULL);
		status = atoi(res);
		if (status != 301)
			fprintf(stderr, "ERROR: SYNC: %s", res);
	} while (status && status != 301);
	if (status != 301)
		return -1;	/* error */
	sleep_usec = (unsigned long) (atof(res + 4) * 1000000);
	if (sleep_usec > 0)
		usleep(sleep_usec);
	return 0;
}

static int sync_mode = 0;

static struct synth_info ext_synth_info[] = {
	{
	 "TiMidity server",	/* name */
	 0,			/* device */
	 SYNTH_TYPE_MIDI,	/* synth_type */
	 SAMPLE_TYPE_BASIC,	/* synth_subtype */
	 0,			/* perc_mode */
	 36,			/* nr_voices */
	 0,			/* nr_drums */
	 128,			/* instr_bank_size */
	 0			/* capabilities */
	 },
	{
	 "TiMidity server",	/* name */
	 1,			/* device */
	 SYNTH_TYPE_SAMPLE,	/* synth_type */
	 SAMPLE_TYPE_AWE32,	/* synth_subtype */
	 0,			/* perc_mode */
	 36,			/* nr_voices */
	 0,			/* nr_drums */
	 128,			/* instr_bank_size */
	 0			/* capabilities */
	 },
};

static double get_time(void)
{
	struct timeval tv;
	struct timezone dmy;

	gettimeofday(&tv, &dmy);

	return (double) tv.tv_sec + (double) tv.tv_usec / 1000000.0;
}

int ioctl_sequencer(struct stSequencer *sequencer, int request, ...)
{
	void *arg;
	va_list ap;
	char *res;

	va_start(ap, request);
	arg = va_arg(ap, void *);
	va_end(ap);

	switch (request) {
	case SNDCTL_SEQ_RESET:
#ifdef	DEBUG
		fprintf(stderr, "ioctl SNDCTL_SEQ_RESET\n");
#endif
		timidity_eot(sequencer);
		timidity_meta_seq(sequencer, 0x01, 0x00, 0);	/* TiMidity reset */
		return 0;

	case SNDCTL_SEQ_SYNC:{
			static double last_time;
			double t;

#ifdef	DEBUG
			fprintf(stderr, "ioctl SNDCTL_SEQ_SYNC\n");
#endif
			if (sync_mode)
				return 0;
			t = get_time();
			if (t - last_time < 0.1)
				return 0;
			last_time = t;
			sync_mode = 1;
			timidity_sync(sequencer,
				      (int) (BUF_HIGH_SYNC * 100));
			sync_mode = 0;
		}
		return 0;

	case SNDCTL_SYNTH_INFO:
#ifdef	DEBUG
		fprintf(stderr, "ioctl SNDCTL_SYNTH_INFO\n");
#endif
		memcpy(arg,
		       &ext_synth_info[((struct synth_info *) arg)->
				       device], sizeof(struct synth_info));
		return 0;

	case SNDCTL_SEQ_CTRLRATE:
#ifdef	DEBUG
		fprintf(stderr, "ioctl SNDCTL_SEQ_CTRLRATE\n");
#endif
		res = timidity_ctl_command(sequencer, "TIMEBASE");
		if (atoi(res) != 200)
			return -1;
		*(int *) arg = atoi(res + 4);
		return 0;


	case SNDCTL_SEQ_NRSYNTHS:
#ifdef	DEBUG
		fprintf(stderr, "ioctl SNDCTL_SEQ_NRSYNTHS\n");
#endif
		*(int *) arg =
		    sizeof(ext_synth_info) / sizeof(ext_synth_info[0]);
		return 0;

	case SNDCTL_SEQ_NRMIDIS:
#ifdef	DEBUG
		fprintf(stderr, "ioctl SNDCTL_SEQ_NRMIDIS\n");
#endif
		*(int *) arg = 1;
		return 0;

	default:
#ifdef	DEBUG
		fprintf(stderr, "ioctl UNKNOWN : %08x\n", request);
#endif
		return -1;
	}
}
#endif
