/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


#include <sys/time.h>
#include <sys/ioctl.h>
#include <sys/soundcard.h>

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>

#include <glib.h>
#include <gnome.h>

#include "timidity-io.h"
#include "my_midi.h"
#include "track.h"
#include "synctmr.h"

static guint lyric_delay = 500; /* 500 ms of delay from music to lyric */

guint track_get_lyric_delay(void) 
{
	return lyric_delay;
}

void track_set_lyric_delay(guint delay)
{
	lyric_delay = delay;
}

typedef struct {
	guint32 dtime;
	guchar status;
	guchar meta_command;
	guchar data1, data2;
	GByteArray *data;
} myevent_t;

typedef struct {
	gint track;
	guint ticks_togo;
	gboolean stopped;
	gint index;
	GPtrArray *events;	/* myevent_t vector */
} track_t;

struct stCurrentFile {
	guint type;
	guint ntracks;
	guint division;		/* ticks per quarter-note */
	GPtrArray *tracks;	/* track_t vector */
};

static myevent_t *event_read(char *data, gint * pos, guchar * last_midi);
static int check_MThd(FILE * file, struct stCurrentFile *current_file);

static long int read_length_size_4(FILE * file)
{
	long int i = 0;
	guchar length[5];
	fread((void *) length, 1, 4, file);
	i = length[3] | (length[2] << 8) | (length[1] << 16) | (length[0]
								<< 24);
	return i;
}

static int isasciistring(char *s)
{
	register int i;
	for (i = 0; i < strlen(s); i++)
		if (s[i] < ' ')
			return FALSE;
	return TRUE;
}

static int read_track(FILE * file, track_t * t)
{				/* Read until we've found a MIDI 1.0 MTrk */
	gint length, pos;
	gchar *data;
	myevent_t *ev;
	guchar last_midi = 0;

	for (;;) {
		unsigned char id[5];
		long int skip;
		fread((void *) id, 1, 4, file);
		id[4] = '\0';
		if (strcmp((char *) id, "MTrk") == 0) {
			break;
		}
		if (isasciistring((char *) id) == FALSE) {
			g_warning("An non-ascii track header found, "
				  "corrupt file.");
			return -1;
		}
		skip = read_length_size_4(file);
		fseek(file, skip, SEEK_CUR);
		if (feof(file))
			return -1;
	}

	length = read_length_size_4(file);
	data = g_malloc(length);
	fread(data, 1, length, file);
	t->stopped = FALSE;
	t->ticks_togo = 0;
	for (pos = 0; pos < length;) {
		ev = event_read(data, &pos, &last_midi);
		g_ptr_array_add(t->events, ev);
	}
	g_free(data);
	return 0;
}

struct stCurrentFile *load_tracks(char *filename)
{
	int i;
	struct stCurrentFile *current_file;
	FILE *file;

	current_file = g_new0(struct stCurrentFile, 1);

	file = fopen(filename, "r");

	if (file == NULL) {
		g_warning("Unnable to open %s.", filename);
		g_free(current_file);
		return NULL;
	}

	if (check_MThd(file, current_file) < 0) {
		g_warning("Invalid header.");
		g_free(current_file);
		fclose(file);
		return NULL;
	}

	current_file->tracks = g_ptr_array_new();

	fseek(file, 14, SEEK_SET);	/* Skip MThd */
	for (i = 0; i < current_file->ntracks; i++) {
		track_t *ptrack;
		ptrack = g_new0(track_t, 1);
		ptrack->track = i;
		g_ptr_array_add(current_file->tracks, ptrack);
		ptrack->events = g_ptr_array_new();
		if (read_track(file, ptrack) < 0) {
			g_warning("Bad track.");
			clear_tracks(current_file);
			fclose(file);
			return NULL;
		}
	}
	fclose(file);
	return current_file;
}

void clear_tracks(struct stCurrentFile *current_file)
{
	gint i, j;
	for (i = 0; i < current_file->tracks->len; i++) {
		track_t *ptrack;
		ptrack = g_ptr_array_index(current_file->tracks, i);
		for (j = 0; j < ptrack->events->len; j++) {
			myevent_t *ev;
			ev = g_ptr_array_index(ptrack->events, j);
			if (ev->data != NULL)
				g_byte_array_free(ev->data, TRUE);
			g_free(ev);
		}
		g_free(ptrack);
	}
	g_ptr_array_free(current_file->tracks, TRUE);
	g_free(current_file);
}

static int check_MThd(FILE * file, struct stCurrentFile *current_file)
{
	char MThd[15];

	fread(MThd, sizeof(char), 14, file);
	/* Checks for Valid MThd Chunk */
	if (strncmp(MThd, "MThd", 4) != 0 || MThd[7] != 6 || MThd[9] > 2)
		return -1;

	current_file->type = MThd[9];
	current_file->ntracks = (unsigned) MThd[11];
	current_file->ntracks += (unsigned) (MThd[10]) << 16;

	if ((MThd[12]) < 0) {
		g_error("File uses SMPTE timing -- Not implemented yet");
		exit(-1);
	}
	current_file->division = (MThd[12] << 8) | (MThd[13] & 0xff);

	return 0;
}

static int read_varlen(char *data, guint * dest)
{
	register int i;
	*dest = 0;
	for (i = 0;; i++) {
		if (data[i] & 0x80) {
			(*dest) |= (long int) (data[i] & 0x7F);
			(*dest) <<= 7;
		} else {
			(*dest) += data[i];
			break;
		}
	}
	if (++i > 4) {		/* Number of bytes read (pos + 1) */
		fprintf(stderr, "malformed variable length %d\n", i);
		exit(-1);
	}
	return i;
}

static myevent_t *event_read(char *data, gint * pos, guchar * last_midi)
{
	myevent_t *ev;
	ev = g_new0(myevent_t, 1);
	*pos += read_varlen(data + *pos, &ev->dtime);

	if ((data[*pos] & 0x80) != 0) {
		ev->status = data[(*pos)++];
	} else {
		if (*last_midi < 0x80) {
			g_warning
			    ("First midi data from a track without status");
			*last_midi = 0x80;
		}
		ev->status = *last_midi;
	}

	ev->meta_command = 0;
	if (ev->status == 0xFF) {
		gint length;
		ev->meta_command = data[(*pos)++];
		*pos += read_varlen(&data[*pos], &length);
		ev->data = g_byte_array_new();
		g_byte_array_append(ev->data, &data[*pos], length);
		*pos += length;
		return ev;
	}
	*last_midi = ev->status;

	if ((ev->status == 0xF0) || (ev->status == 0xF7)) {
		gint length;
		*pos += read_varlen(&data[*pos], &length);
		ev->data = g_byte_array_new();
		g_byte_array_append(ev->data, &data[*pos], length);
		*pos += length;
		return ev;
	}

	switch (ev->status & 0xF0) {
	case NOTE_ON:
	case NOTE_OFF:
	case PITCH_BEND:
	case CONTROL_CHANGE:
	case KEY_PRESSURE:
		ev->data1 = data[(*pos)++];
		ev->data2 = data[(*pos)++];
		break;
	case PROGRAM_CHANGE:
	case CHANNEL_PRESSURE:
		ev->data1 = data[(*pos)++];
		break;
	}
	return ev;
}

static void seq_playevent(struct stPlayHandle *playhandle,
			  track_t * ptrack, myevent_t * event);

static void send_midi_block(double ms_available, gpointer data);

struct stPlayHandle {
	struct stSequencer *sequencer;
	struct stCurrentFile *current_file;
	void (*define_lyric_line) (gint line, gchar * left, gchar * right);
	void (*change_background_image) (void);
	gboolean playing, end_of_lyric;
	guint position_ms;
	guint tempo;		/* microseconds per quarter-note */
	gdouble ms_per_tick;
	struct stSyncTmr *synctmr;
	GString *lyric_past, *lyric_next, *lyric_nextline;
	gint lyric_line;
};

static gchar *try_to_get_next_two_lyric_lines(gchar * alreadyhave,
					      track_t * ptrack);
static void lyric_event_new_line(struct stPlayHandle *playhandle,
				 gchar * nextlyrics);

struct stPlayHandle *track_playfile(struct stSequencer *sequencer,
				    struct stCurrentFile *current_file,
				    void (*define_lyric_line) (gint line,
							       gchar *
							       left,
							       gchar *
							       right),
				    void (*change_background_image) (void))
{
	int i;
	track_t *ptrack;
	struct stPlayHandle *playhandle;

	playhandle = g_new(struct stPlayHandle, 1);
	playhandle->sequencer = sequencer;
	playhandle->current_file = current_file;
	playhandle->define_lyric_line = define_lyric_line;
	playhandle->change_background_image = change_background_image;
	playhandle->playing = TRUE;
	playhandle->end_of_lyric = FALSE;

	playhandle->synctmr = synctmr_create();
	playhandle->lyric_past = g_string_new("");
	playhandle->lyric_next = g_string_new("");
	playhandle->lyric_nextline = g_string_new("");
	playhandle->lyric_line = 0;

	for (i = 0; i < current_file->ntracks; i++) {
		myevent_t *event;
		gchar *text;
		ptrack = g_ptr_array_index(current_file->tracks, i);
		ptrack->index = 0;
		event = g_ptr_array_index(ptrack->events, ptrack->index);
		ptrack->ticks_togo = event->dtime;
		text = try_to_get_next_two_lyric_lines("", ptrack);
		if ((strchr(text, '/') != NULL)
		    || (strchr(text, '\\') != NULL)) {
			lyric_event_new_line(playhandle, text);
		}
		g_free(text);
	}

	timidity_set_sync_cb(playhandle->sequencer,
			     send_midi_block, playhandle);
	return playhandle;
}

static gint end_of_play_and_lyric(gpointer data)
{
	struct stPlayHandle *playhandle = data;
	if (playhandle->end_of_lyric == FALSE) {
		synctmr_destroy(playhandle->synctmr);	/* call the pending timers */
		(*playhandle->define_lyric_line) (0, "", "");
		(*playhandle->define_lyric_line) (1, "", "");
		playhandle->end_of_lyric = TRUE;
		g_string_free(playhandle->lyric_past, TRUE);
		g_string_free(playhandle->lyric_next, TRUE);
		g_string_free(playhandle->lyric_nextline, TRUE);
	}
	return FALSE;
}

static void send_midi_block(double ms_available, gpointer data)
{
	struct stPlayHandle *playhandle = data;
	int i;
	track_t *ptrack;
	guint lowest;
	double blockavail_ms, lowest_ms, position_ms;
	if (!playhandle->playing) {
		sequencer_wait(playhandle->sequencer, ms_available);
		return;
	}

	blockavail_ms = ms_available;
	position_ms = ms_available + lyric_delay;
	while (blockavail_ms > 0) {
		playhandle->position_ms = (guint) position_ms;
		lowest = 0x0FFFFFFF;
		for (i = 0; i < playhandle->current_file->ntracks; i++) {
			ptrack =
			    (track_t *) g_ptr_array_index(playhandle->
							  current_file->
							  tracks, i);
			if (ptrack->stopped)
				continue;
			while (ptrack->ticks_togo == 0) {
				myevent_t *event;
				event = (myevent_t *)
				    g_ptr_array_index(ptrack->events,
						      ptrack->index);
				seq_playevent(playhandle, ptrack, event);
				ptrack->index++;
				if (ptrack->index >= ptrack->events->len) {
					ptrack->stopped = TRUE;
					break;
				}
				event = (myevent_t *)
				    g_ptr_array_index(ptrack->events,
						      ptrack->index);
				ptrack->ticks_togo = event->dtime;
			}
			if (ptrack->stopped)
				continue;
			if (lowest > ptrack->ticks_togo)
				lowest = ptrack->ticks_togo;
		}

		if (lowest == 0x0FFFFFFF) {	/* end of the music */
			synctmr_timeout_add(playhandle->synctmr,
					    playhandle->position_ms,
					    end_of_play_and_lyric,
					    playhandle);
			playhandle->playing = FALSE;
			return;
		}

		lowest_ms = lowest * playhandle->ms_per_tick;
		if (lowest_ms > blockavail_ms) {
			lowest_ms = blockavail_ms;
			lowest =
			    (guint) (lowest_ms / playhandle->ms_per_tick);
			blockavail_ms = 0.0;
		} else {
			blockavail_ms -= lowest_ms;
			position_ms += lowest_ms;
		}
		for (i = 0; i < playhandle->current_file->ntracks; i++) {
			ptrack =
			    g_ptr_array_index(playhandle->current_file->
					      tracks, i);
			ptrack->ticks_togo -= lowest;
		}
		sequencer_wait(playhandle->sequencer, lowest_ms);
	}
}

void track_clear_playfile(struct stPlayHandle
			  *playhandle)
{
	timidity_set_sync_cb(playhandle->sequencer, NULL, NULL);
	end_of_play_and_lyric(playhandle);
	g_free(playhandle);
}

static void lyric_event(struct stPlayHandle
			*playhandle, track_t * ptrack, myevent_t * event);
static void seq_playevent(struct stPlayHandle
			  *playhandle, track_t * ptrack, myevent_t * event)
{
	if (event->status != 0xff) {	/* it's a midi event */
		sequencer_send_midi_event(playhandle->sequencer,
					  event->status, event->data1,
					  event->data2);
		return;
	}

	/* it's a metaevent */
	switch (event->meta_command) {
	case METACMD_TRACK_NAME:
		{
			gchar *trackname;
			trackname =
			    g_strndup(event->data->data, event->data->len);
			g_free(trackname);
		}
		break;
	case METACMD_TEXT_EV:
		lyric_event(playhandle, ptrack, event);
		break;
	case METACMD_EOT:
		ptrack->stopped = TRUE;
		break;
	case METACMD_SETTEMPO:
		{
			unsigned char s[3];
			if (event->data->len < 3)
				break;
			strncpy((char *) s, event->data->data, 3);
			playhandle->tempo =
			    (s[0] << 16) | (s[1] << 8) | s[2];
			playhandle->ms_per_tick =
			    playhandle->tempo /
			    (playhandle->current_file->division * 1000.0);
		}
		break;
	}
}

static gboolean try_to_find_next_lyric_event(track_t *
					     ptrack,
					     gint *
					     index, myevent_t ** newevent)
{
	myevent_t *event;
	*newevent = NULL;
	if (*index < 0)
		return FALSE;
	for (;;) {
		(*index)++;
		if (*index >= ptrack->events->len)
			return FALSE;
		event =
		    (myevent_t *) g_ptr_array_index(ptrack->events,
						    *index);
		if ((event->status == 0xff)
		    && (event->meta_command == METACMD_TEXT_EV)) {
			*newevent = event;
			return TRUE;
		}
	}
}

static gchar *try_to_get_next_two_lyric_lines(gchar *
					      alreadyhave,
					      track_t * ptrack)
{
	GString *lines;
	gchar *res;
	gint index;
	index = ptrack->index;
	lines = g_string_new(alreadyhave);
	for (;;) {
		myevent_t *newevent;
		gchar *ps;
		gint cont = 0;
		gchar *buf;
		for (ps = lines->str; *ps != '\0'; ps++) {
			if ((*ps == '/') || (*ps == '\\')) {
				*ps = '/';
				cont++;
				if (cont == 2)
					break;
			}
		}
		if (cont >= 2)
			break;
		if (try_to_find_next_lyric_event(ptrack, &index, &newevent)
		    == FALSE)
			break;
		if (newevent->data->len > 0) {
			buf =
			    g_strndup(newevent->data->data,
				      newevent->data->len);
			if (buf[0] != '@')
				g_string_append(lines, buf);
			g_free(buf);
		}
	}
	res = lines->str;
	g_string_free(lines, FALSE);
	return res;
}

static void lyric_event_text_only(struct stPlayHandle
				  *playhandle, gchar * text)
{
	g_string_append(playhandle->lyric_past, text);
	g_string_erase(playhandle->lyric_next, 0,
		       MIN(strlen(text), playhandle->lyric_next->len));
	(*playhandle->define_lyric_line) (playhandle->
					  lyric_line,
					  playhandle->
					  lyric_past->str,
					  playhandle->lyric_next->str);
}

static void lyric_event_new_line(struct stPlayHandle
				 *playhandle, gchar * nextlyrics)
{
	gchar **lines;
	g_string_truncate(playhandle->lyric_past, 0);
	g_string_truncate(playhandle->lyric_next, 0);
	g_string_truncate(playhandle->lyric_nextline, 0);
	lines = g_strsplit(nextlyrics, "/", 3);
	if (lines[0] != NULL) {
		g_string_assign(playhandle->lyric_next, lines[0]);
		if (lines[1] != NULL) {
			g_string_assign(playhandle->lyric_nextline,
					lines[1]);
		}
	}
	g_strfreev(lines);
	(*playhandle->define_lyric_line) (playhandle->
					  lyric_line, "",
					  playhandle->lyric_nextline->str);
	playhandle->lyric_line = (playhandle->lyric_line == 1 ? 0 : 1);
	(*playhandle->define_lyric_line) (playhandle->
					  lyric_line, "",
					  playhandle->lyric_next->str);
}

struct stTextEvent {
	struct stPlayHandle *playhandle;
	enum { EV_TEXT_ONLY, EV_NEW_LINE, EV_NEW_SCREEN } type;
	gchar *text;
};

static gint lyric_event_sync(gpointer data)
{
	struct stTextEvent *evtext = data;
	if (evtext->type == EV_TEXT_ONLY) {
		lyric_event_text_only(evtext->playhandle, evtext->text);
	} else {
		if(evtext->type == EV_NEW_SCREEN)
			(*evtext->playhandle->change_background_image)();
		lyric_event_new_line(evtext->playhandle, evtext->text);
	}
	g_free(evtext->text);
	g_free(evtext);
	return FALSE;
}

static void lyric_event(struct stPlayHandle
			*playhandle, track_t * ptrack, myevent_t * event)
{
	gchar *text, *pos;
	if (event->data->len == 0)
		return;
	if (event->data->data[0] == '@')
		return;
	text = g_strndup(event->data->data, event->data->len);
	for (pos = text; *pos != '\0';) {
		gchar *refpos = pos;
		for (pos = refpos; *pos != '\0'; pos++) {
			if ((*pos == '\\') || (*pos == '/')) {
				break;
			}
		}
		if (pos != refpos) {
			struct stTextEvent *evtext;
			evtext = g_new(struct stTextEvent, 1);
			evtext->playhandle = playhandle;
			evtext->type = EV_TEXT_ONLY;
			evtext->text = g_strndup(refpos, pos - refpos);
			synctmr_timeout_add(playhandle->synctmr,
					    playhandle->position_ms,
					    lyric_event_sync, evtext);
		}
		if (*pos != '\0') {
			struct stTextEvent *evtext;
			evtext = g_new(struct stTextEvent, 1);
			evtext->playhandle = playhandle;
			evtext->type =
			    (*pos == '\\' ? EV_NEW_SCREEN : EV_NEW_LINE);
			evtext->text =
			    try_to_get_next_two_lyric_lines(pos + 1,
							    ptrack);
			synctmr_timeout_add(playhandle->synctmr,
					    playhandle->position_ms,
					    lyric_event_sync, evtext);
			pos++;
		}
	}
	g_free(text);
}
